function y_cum = cumest(y,norder,maxlag,nsamp,overlap,flag,k1,k2)
%CUMEST	Second-, third- or fourth-order cumulants.
%	y_cum = cumest (y, norder, maxlag, samp_seg, overlap, flag, k1, k2)
%	       y - time-series  - should be a vector
%	  norder - cumulant order: 2, 3 or 4 [default = 2]
%	  maxlag - maximum cumulant lag to compute [default = 0]
%	samp_seg - samples per segment  [default = data_length]
%	 overlap - percentage overlap of segments [default = 0]
%	           overlap is clipped to the allowed range of [0,99].
%	   flag  - 'biased' or 'unbiased'  [default = 'biased']
%	  k1,k2  - specify the slice of 3rd or 4th order cumulants
%	  y_cum  - C2(m) or C3(m,k1) or C4(m,k1,k2),  -maxlag <= m <= maxlag
%	           depending upon the cumulant order selected

% Functions cumN_est (N=2,3,4) should be invoked
%  via this routine, since this routine does extensive parameter checking.

%  Copyright (c) 1991-98 by United Signals & Systems, Inc. and The MathWorks, Inc.
%       $Revision: 1.3 $
%  A. Swami, January 20, 1993;  August 6, 1994.

%     RESTRICTED RIGHTS LEGEND
% Use, duplication, or disclosure by the Government is subject to
% restrictions as set forth in subparagraph (c) (1) (ii) of the
% Rights in Technical Data and Computer Software clause of DFARS
% 252.227-7013.
% Manufacturer: United Signals & Systems, Inc., P.O. Box 2374,
% Culver City, California 90231.
%
%  This material may be reproduced by or for the U.S. Government pursuant
%  to the copyright license under the clause at DFARS 252.227-7013.

% ----------- Parameter checks ------------------

 [ksamp,nrecs] = size(y);
 if (ksamp == 1)
     ksamp = nrecs; nrecs = 1;
 end
 if (exist('norder') ~= 1) norder = 2; end
 if (norder < 2 | norder > 4)
    error('cumulant order must be 2, 3 or 4')
 end
 if (exist('maxlag') ~= 1) maxlag = 0; end
 if (maxlag < 0)
    error ('"maxlag" must be non-negative ')
 end

 if (exist('nsamp') ~= 1) nsamp = ksamp; end
 if (nrecs > 1)           nsamp = ksamp; end
 if (nsamp <= 0 | nsamp > ksamp)
    nsamp = ksamp;
 end

 if (exist('overlap') ~= 1) overlap = 0; end
 if (nrecs > 1)             overlap = 0; end
 overlap = max(0,min(overlap,99));

 if (exist('flag') ~= 1) flag = 'biased'; end
 if (flag(1:1) ~= 'b' & flag(1:1) ~= 'B')   flag = 'unbiased';
 else, flag = 'biased';
 end

 if (exist('k1') ~= 1) k1 = 0; end
 if (exist('k2') ~= 1) k2 = 0; end

% ---------- go, estimate the cumulants ------------------------
 if (norder == 2)
    y_cum = cum2est (y, maxlag, nsamp, overlap, flag);

 elseif (norder == 3)
    y_cum = cum3est (y, maxlag, nsamp, overlap, flag, k1);

 elseif (norder == 4)
    y_cum = cum4est (y, maxlag, nsamp, overlap, flag, k1, k2);

 end

return


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Second Order Cumulant
function   y_cum = cum2est (y, maxlag, nsamp, overlap, flag)
%CUM2EST Covariance function.
%	Should be involed via "CUMEST" for proper parameter checks.
%	y_cum = cum2est (y, maxlag, samp_seg, overlap,  flag)

%	       y: input data vector (column)
%	  maxlag: maximum lag to be computed
%	samp_seg: samples per segment (<=0 means no segmentation)
%	 overlap: percentage overlap of segments
%	    flag: 'biased', biased estimates are computed
%	          'unbiased', unbiased estimates are computed.
%	   y_cum: estimated covariance,
%	          C2(m)  -maxlag <= m <= maxlag
%	all parameters must be specified!


%  Copyright (c) 1991-98 by United Signals & Systems, Inc. and The MathWorks, Inc.
%       $Revision: 1.4 $
%  A. Swami   January 20, 1993


%     RESTRICTED RIGHTS LEGEND
% Use, duplication, or disclosure by the Government is subject to
% restrictions as set forth in subparagraph (c) (1) (ii) of the
% Rights in Technical Data and Computer Software clause of DFARS
% 252.227-7013.
% Manufacturer: United Signals & Systems, Inc., P.O. Box 2374,
% Culver City, California 90231.
%
%  This material may be reproduced by or for the U.S. Government pursuant
%  to the copyright license under the clause at DFARS 252.227-7013.

% C2(m) := E conj(x(n)) x(n+k)

% ----------  parameter checks are done by CUMEST  ----------------

   [n1,n2] = size(y);    N = n1*n2;

   overlap  = fix(overlap/100 * nsamp);
   nrecord  = fix( (N - overlap)/(nsamp - overlap) );
   nadvance = nsamp - overlap;

   y_cum    = zeros(maxlag+1,1);
   ind = 1:nsamp;

   for i=1:nrecord
       x = y(ind); x = x(:) - mean(x);     % make sure we have a colvec
       for k = 0:maxlag
           y_cum(k+1) = y_cum(k+1) + x([1:nsamp-k])' * x([k+1:nsamp]);
       end
       ind = ind + nadvance;
   end
   if (flag(1:1) == 'b' | flag(1:1) == 'B')
       y_cum = y_cum / (nsamp*nrecord);
   else
       y_cum = y_cum ./ (nrecord * (nsamp-[0:maxlag]' ));
   end
   if maxlag > 0,
      y_cum = [conj(y_cum(maxlag+1:-1:2)); y_cum];
   end

return



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Third order cumulant
function   y_cum = cum3est (y, maxlag, nsamp, overlap, flag, k1)
%CUM3EST Third-order cumulants.
%	Should be invoked via "CUMEST" for proper parameter checks
%	y_cum = cum3est (y, maxlag, samp_seg, overlap, flag, k1)

%	y_cum = cum3est (y, maxlag, samp_seg, overlap, flag, k1)
%	       y: input data vector (column)
%	  maxlag: maximum lag to be computed
%	samp_seg: samples per segment
%	 overlap: percentage overlap of segments
%	   flag : 'biased', biased estimates are computed  [default]
%	          'unbiased', unbiased estimates are computed.
%	      k1: the fixed lag in c3(m,k1): see below
%	   y_cum:  estimated third-order cumulant,
%	           C3(m,k1)  -maxlag <= m <= maxlag

%  Copyright (c) 1991-98 by United Signals & Systems, Inc. and The MathWorks, Inc.
%       $Revision: 1.3 $
%  A. Swami   January 20, 1993

% Modified Jan 20, 94 to handle complex case properly.
%  c3(i,j) := E x^*(n) x(n+i) x(n+j)   (x assumed zero mean)

%     RESTRICTED RIGHTS LEGEND
% Use, duplication, or disclosure by the Government is subject to
% restrictions as set forth in subparagraph (c) (1) (ii) of the
% Rights in Technical Data and Computer Software clause of DFARS
% 252.227-7013.
% Manufacturer: United Signals & Systems, Inc., P.O. Box 2374,
% Culver City, California 90231.
%
%  This material may be reproduced by or for the U.S. Government pursuant
%  to the copyright license under the clause at DFARS 252.227-7013.

%  c3(i,j) := E x^*(n) x(n+i) x(n+j)   (x assumed zero mean)

%---------------- Parameter checks done by CUMEST --------------
   [n1,n2]  = size(y);
   N        = n1*n2;
   minlag   = -maxlag;
   overlap  = fix(overlap/100 * nsamp);
   nrecord  = fix( (N - overlap)/(nsamp - overlap) );
   nadvance = nsamp - overlap;

   y_cum = zeros(maxlag-minlag+1,1);

   ind = (1:nsamp)';
   nlags = 2 * maxlag + 1;
   zlag  = 1 + maxlag;
   if (flag(1) == 'b' | flag(1) == 'B')
   	scale = ones(nlags,1)/nsamp;
   else
       lsamp = nsamp - abs(k1);
       scale = [lsamp-maxlag:lsamp,lsamp-1:-1:lsamp-maxlag]';
       [m2,n2] = size(scale);
       scale = ones(m2,n2) ./ scale;
   end

   for i=1:nrecord
       x = y(ind); x = x(:) - mean(x);     % make sure we have a col vec
       cx = conj(x);
       z = x*0;

%                     create the "IV" matrix: offset for second lag

       if (k1 >= 0) z(1:nsamp-k1)  = x(1:nsamp-k1,:) .* cx(k1+1: nsamp,:);
       else         z(-k1+1:nsamp) = x(-k1+1:nsamp)  .* cx(1:nsamp+k1);
       end

%                     compute third-order cumulants

       y_cum(zlag)  =  y_cum(zlag) + z' * x;

       for k = 1:maxlag
           y_cum(zlag-k) = y_cum(zlag-k) + z([k+1:nsamp])' * x([1:nsamp-k]);
           y_cum(zlag+k) = y_cum(zlag+k) + z([1:nsamp-k])' * x([k+1:nsamp]);
       end

       ind = ind + nadvance;
   end

   y_cum = y_cum .* scale / nrecord;

return


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  Fourth order cumulant
function   y_cum = cum4est (y, maxlag, nsamp, overlap, flag, k1, k2)
%CUM4EST Fourth-order cumulants.
%       Should be invoked via CUMEST for proper parameter checks
%       y_cum = cum4est (y, maxlag, samp_seg, overlap, flag, k1, k2)

%       Computes sample estimates of fourth-order cumulants
%       via the overlapped segment method.
%
%       y_cum = cum4est (y, maxlag, samp_seg, overlap, flag, k1, k2)
%              y: input data vector (column)
%         maxlag: maximum lag
%       samp_seg: samples per segment
%        overlap: percentage overlap of segments
%          flag : 'biased', biased estimates are computed
%               : 'unbiased', unbiased estimates are computed.
%	  k1,k2 : the fixed lags in C3(m,k1) or C4(m,k1,k2); see below
%	  y_cum : estimated fourth-order cumulant slice
%	          C4(m,k1,k2)  -maxlag <= m <= maxlag
% 	Note: all parameters must be specified


%  Copyright (c) 1991-98 by United Signals & Systems, Inc. and The MathWorks, Inc.
%       $Revision: 1.3 $
%  A. Swami   January 20, 1993

% Modified, Januar 20, 1994 to handle complex case properly:
% c4(t1,t2,t3) := cum( x^*(t), x(t+t1), x(t+t2), x^*(t+t3) )

%     RESTRICTED RIGHTS LEGEND
% Use, duplication, or disclosure by the Government is subject to
% restrictions as set forth in subparagraph (c) (1) (ii) of the
% Rights in Technical Data and Computer Software clause of DFARS
% 252.227-7013.
% Manufacturer: United Signals & Systems, Inc., P.O. Box 2374,
% Culver City, California 90231.
%
%  This material may be reproduced by or for the U.S. Government pursuant
%  to the copyright license under the clause at DFARS 252.227-7013.


% c4(t1,t2,t3) := cum( x^*(t), x(t+t1), x(t+t2), x^*(t+t3) )
%  cum(w,x,y,z) := E(wxyz) - E(wx)E(yz) - E(wy)E(xz) - E(wz)E(xy)
%  and, w,x,y,z are assumed to be zero-mean.


% ---- Parameter checks are done in CUMEST ----------------------
   [n1,n2]  = size(y);
   N        = n1 * n2;
   overlap0 = overlap;
   overlap  = fix(overlap/100 * nsamp);
   nrecord  = fix( (N - overlap)/(nsamp - overlap) );
   nadvance = nsamp - overlap;


% ------ scale factors for unbiased estimates --------------------

   nlags = 2 * maxlag + 1;
   zlag  = 1 + maxlag;
   tmp   = zeros(nlags,1);
   if (flag(1:1) == 'b'  | flag(1:1) == 'B')
       scale = ones(nlags,1) / nsamp;
   else
       ind   = [-maxlag:maxlag]';
       kmin  = min(0,min(k1,k2));
       kmax  = max(0,max(k1,k2));
       scale = nsamp - max(ind,kmax) + min(ind,kmin);
       scale = ones(nlags,1) ./ scale;
   end
   mlag  = maxlag + max(abs([k1,k2]));
   mlag  = max( mlag, abs(k1-k2) );
   mlag1 = mlag + 1;
   nlag  = maxlag;
   m2k2  = zeros(2*maxlag+1,1);

   if (any(any(imag(y) ~= 0))) complex_flag = 1;
   else complex_flag = 0;
   end

% ----------- estimate second- and fourth-order moments; combine ------

   y_cum  = zeros(2*maxlag+1,1);
   R_yy   = zeros(2*mlag+1,1);

   ind   = 1:nsamp;
   for i=1:nrecord
       tmp = y_cum * 0 ;
       x = y(ind); x = x(:) - mean(x);  z =  x * 0;  cx = conj(x);
%                     create the "IV" matrix: offset for second lag

       if (k1 >= 0)
       		z(1:nsamp-k1)  = x(1:nsamp-k1,:) .* cx(k1+1: nsamp,:);
       else
       		z(-k1+1:nsamp) = x(-k1+1:nsamp)  .* cx(1:nsamp+k1);
       end

%                     create the "IV" matrix: offset for third lag

       if (k2 >= 0)
          z(1:nsamp-k2) = z(1:nsamp-k2) .* x(k2+1: nsamp);
          z(nsamp-k2+1:nsamp) = zeros(k2,1);
       else
          z(-k2+1:nsamp) = z(-k2+1:nsamp) .* x(1:nsamp+k2);
          z(1:-k2)    = zeros(-k2,1);
       end

       tmp(zlag)  =  tmp(zlag) + z' * x;
       for k = 1:maxlag
           tmp(zlag-k) = tmp(zlag-k) + z([k+1:nsamp])' * x([1:nsamp-k]);
           tmp(zlag+k) = tmp(zlag+k) + z([1:nsamp-k])' * x([k+1:nsamp]);
       end

       y_cum = y_cum + tmp .* scale ;

       R_yy = cum2est(x,mlag,nsamp,overlap0,flag);
       if (complex_flag)    % We need E x(t)x(t+tau) stuff also:
       	   M_yy  = cum2x(conj(x),x,mlag,nsamp,overlap0,flag);
       else
           M_yy  = R_yy;
       end
       y_cum = y_cum ...
           - R_yy(mlag1+k1) * R_yy(mlag1-k2-nlag:mlag1-k2+nlag) ...
           - R_yy(k1-k2+mlag1) * R_yy(mlag1-nlag:mlag1+nlag)  ...
           - M_yy(mlag1+k2)' * M_yy(mlag1-k1-nlag:mlag1-k1+nlag) ;

       ind = ind + nadvance;
end

y_cum = y_cum / nrecord;
return


