function [H,Vmodel]=vkernels(x,y,P,M);
%VKERNELS	Second-order Volterra system identification
% [h, H] = VKERNELS (x, y, P, M);
% INPUT 
%        x - input of a system (a column vector)
%        y - output of a system (a column vector)
%        P - nonlinearity order
%        M - assumed system memory (in samples). Could be a vector
%            if different memory lengths are assumend for different nonlinearities
% OUTPUT
%        H - cell array of kernels up to order N
%   Vmodel.* - structure containing the trained model
%        *.h - vector of kernel coefficients (symmetric kernel)
%        *.M - model memory ( a vector )
%        *.P - nonlinearity order 


%  Rev 1.0    Dec. 07, 2003,  Dmitriy Shutin  <dshutin@inw.tugraz.at>
%                             doesn't work yet for complex signals

% Parameter checks ------------------------------------------------
if (exist('x') ~= 1  | exist('y') ~= 1)
   error('both x and y must be specified')
end
x=x(:);
y=y(:);
if (size(x) ~= size(y))
   error('x and y must have the same dimensions')
end
if ~exist('P')
    P=2;
end    
 
if ~exist('M')
    M=4*ones(P,1);
else
    if length(M)==1    
        M=ones(P,1)*M;
    end        
end


% computing required matrices  Ah=y -------------------------------
nsamp = size(x,1);

% data matrix
A=[];
for k=1:P
    X=[zeros( M(k)-1,1) ; x ];
    [R, Ind{k}]=constr_mat(X,k,M(k));
    A=[A, R];
end  
% accounting for a DC
A=[ones(size(A,1),1),A];
  
  
%checking the invertability
fprintf(1,'Matrix COND=%d\n',cond(A));
% solving for coefficients
h=pinv(A)*y;

% Creating kernel matrices   --------------------------------------

% temporal variable for operations
q=h;
% zero order kernel
H{1}=q(1);
q(1)=[];

for k=1:P
    %extracting kernel indices
    ind=Ind{k};
    q_vec=q( 1:size(ind,1) );
    if (k>1 & M(k)>1)
        Q=zeros(M(k)*ones(1,k));       
        % only real inputs are supported
        if isreal(h) | 1
        % in case h is real, we fill in symmetric entries -  h(i,j)=h(j,i)
        % all possible permutations of the index matrix 
            Perm=perms(1:k);
            [NPerm]=size(Perm,1);
            for l=1:NPerm
                Pind=ind( : ,Perm(l,:) );
                % converting subscripts into linear index (a la sub2ind)
                [ndx]=convSub2Ind(size(Q),Pind);
                Q(ndx)=q_vec;
            end
        else
            % if matrix is complex, no symmetry is assumed.
            
            % converting subscripts into linear index (a la sub2ind)
            [ndx]=convSub2Ind(size(Q),ind);
            Q(ndx)=q_vec;
        end %is real
    else
        Q=zeros(M(k),1);
        Q(ind)=q_vec;
    end % if k>1  
    H{k+1}=Q;
    q( 1:size(ind,1) )=[];
end

Vmodel.h=h;
Vmodel.M=M;
Vmodel.P=P;



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       Converting subscript to linear index 
%              (a la sub2ind but for a matrix)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ndx]=convSub2Ind(siz,ind)
K = [1 cumprod(siz(1:end-1))];
ndx = 1;
for i = 1:size(ind,2),
    ndx = ndx + (ind(:,i)-1)*K(i);
end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%       Construction of the data matrix 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [R,Ind]=constr_mat(x,N,mem)
% construction of data matrix A for LS solving for kernel coeffficients 
%   x - input signal
%   N - nonlinearity order
% mem - memory length
 
% constructing the index vector 
% Only real inputs are supported
if isreal(x) | 1
    % computing indices ( symmetric kernel)
    [Ind]=create_ind(mem,N,1);
else
    % computing indices ( a-symmetric kernel (complex), full permutations)
    ind=[1:mem];
    Ind=ind;
    for k=2:N
        Ind=combvec(ind,Ind);
    end
    Ind=Ind';
end

X = buffer(x, mem, mem-1,'nodelay');
[mX]=size(X,2);
R=[];
for k=1:mX
    d=X(end:-1:1,k);
    R=[R; prod( d(Ind),2 )'];
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%         Constructing index for real signals
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [Out]=create_ind(mem,N,stInd)
%  recurrent index computation using kernel symmetry 
%   mem - memory length
%     N - nonlinearity order
% stInd - start index

%  The idea is based on the reccurent formalization of such loop structures 
%   for i=1:N
%      for j=i:N
%        for k=j:N 
%            ...  

Out=[];

if N==1
    Out=[stInd:mem]';
    return
else
    for k=stInd:mem
        [out]=create_ind(mem,N-1,k);
        Add=[k*ones(size(out,1),1), out ];
        Out=[Out; Add];
    end
end
